# toyMS2.R
#
# Linear VAR DGP to show that forecast evaluation will get the correct
# model, and that it is not just an MS thing.
#
#
# Patrick T. Brandt
#
#
# 20120228 : Initial version (based on toymix.R from POLMETH 2011)
# 20120613 : Updated to work with latest MSBVAR devel and to
#            empirically compute the PITs
# 20120719 : Run another analysis for the MSBVAR updates.
# 20130920 : Corrected comments on the DGP
# 20130921 : Add more comments what the script does.
# 20130927 : Version to do linear VAR

# Load libraries
library(MSBVAR)

# Set sample size
TT <- 550

# Set parameters
m <- 2  # number of endogenous variables
p <- 1  # lag length


# N(-1, 0.5) v. N(1, 0.5)...  average out to N(0, 0.5)... but are
# distinct, but made very simple.

# VAR param
intercepts  <- c(-1,1)  # intercepts
ARs  <- matrix(c(.7, .1,
                 .7, .1), m, byrow=TRUE)

# Variance-Covariance Matrix for
# MVN distributed disturbances
Sigma  <- 0.5*diag(m)


# Generate the simulated VAR data

Y <- matrix(0, TT, 2)
set.seed(81571)
for(i in 1:TT)
{
    # draw errors
    e <- rmvnorm(1, sigma=Sigma)

    if(i==1) {
        Y[1,] <- ARs%*%matrix(c(rnorm(1, -1, sqrt(0.5)),
                                rnorm(1, 1, sqrt(0.5))),
                                2,1) + t(e)
    } else {
        Y[i,] <- ARs%*%(Y[i-1,]) + (intercepts) + t(e)
    }
}


# Set training set and ex ante forecast horixon
Y <- ts(Y)
Y.past <- window(Y, start=1, end=500)
Y.future <- window(Y, start=501, end=520)

# Plot data
pdf(file="toyMS2data.pdf", width=8, height=6)

plot(window(Y, end=c(520)), plot.type="s", lty=1:2, main="Data Series",
     ylab="", xlab="")
abline(h=0)
abline(h=3, lty=2)
abline(h=-3, lty=2)
abline(v=500, lty=2)

dev.off()

################################################
# Estimate models, based on the training set
################################################

# Estimate posterior mode under a diffuse prior
set.seed(123)
M0 <- msbvar(Y.past, p=1, h=2, lambda0=0.5, lambda1=1, lambda3=1,
             lambda4=0.5, lambda5=1, mu5=0, mu6=0, qm=4,
             alpha.prior=matrix(c(10,4,4,10), 2, 2), prior=0,
             max.iter=100)


# Now plot the EM regime probs
plot(ts(M0$fp),
     plot.type="s", col=1:3, ylab="Regime probailities")

# Gibbs sampler the posterior
N1 <- 1000
N2 <- 5000

# Now fit the posterior density.
set.seed(2317)

M0.gibbs <- gibbs.msbvar(M0, N1, N2, permute=TRUE, Q.method="Gibbs")

plotregimeid(M0.gibbs, type="intercepts", ask=FALSE)

# Now sample with identification
M0id.gibbs <- gibbs.msbvar(M0, N1, N2,
                           permute=FALSE, Beta.idx=c(3,1))


# Get summaries
summary(M0id.gibbs$Q.sample)
summary(M0id.gibbs$Beta.sample)
summary(M0id.gibbs$Sigma.sample)

plot.SS(M0id.gibbs)

###################################################
# Forecast
#
# 1) DGP forecast based on MS model
# 2) Normal linear projection (no switching) from a VAR
# 3) Incorrect N(0,1)

#source("~/MSBVARlocal/devel-unstable/msbvarforecast.R")
N1 <- 100
N2 <- 1000
k <- 20
system.time(ftmp <- forecast(M0.gibbs, k, N1=N1, N2=N2))

# Now summaries of forecasts
fout <- apply(ftmp$forecast, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))

rd <- Y.future[1:k,]

par(mfrow=c(2,1))
plot(ts(cbind(rd[,1],t(fout[,,1]))), plot.type="s", col=c(1,2,2,2),
     lty=c(1,2,1,2), ylab="Series 1")
plot(ts(cbind(rd[,2],t(fout[,,2]))), plot.type="s", col=c(1,2,2,2),
     lty=c(1,2,1,2), ylab="Series 2")


################################################################
# Other forecasts, based on the sample
#
# 1) VAR (no switching)
# 2) Sample means & variances
################################################################

# VAR forecast
VAR1 <- szbvar(Y.past, p=1, lambda0=1, lambda1=0.1, lambda3=1,
               lambda4=0.25, lambda5=0, mu5=0,
               mu6=0, qm=4, prior=0)

fVAR <- forecast(VAR1, nsteps=k)

fVARuc <- uc.forecast(VAR1, nsteps=k, burnin=N1, gibbs=N2)
fVAR <- fVARuc$forecast

#############################################################
# Incorrect sample mean forecasts
#############################################################
mn <- apply(Y.past, 2, mean)
covs <- cov(Y.past)

tmp <- rmvnorm(m*k*N2, mean=mn, sigma=covs)
normonly <- array(0, c(N2, k, m))
normonly[,,1] <- array(tmp[,1], c(N2, k, 1))
normonly[,,2] <- array(tmp[,2], c(N2, k, 1))

#############################################################
# Compute the summary stats over the sample of forecasts:
# 1) Average RMSE
# 2) Average MAE
# 3) Average Gaussian approx.
# 4) Empirical CDF
#############################################################
TT <- nrow(Y.past)
MS <- ftmp$forecast
VAR <- fVAR

#############################################################
# Average RMSEs / MAE / over k forecasts
#############################################################
rmse.VAR <- rmse.norm <- rmse.MS <- matrix(0, 1, 2)
mae.VAR <- mae.norm <- mae.MS <- matrix(0, 1, 2)
for(i in 1:m)
{
    # RMSE
    rmse.MS[1,i] <- mean(sqrt(apply((MS[,,i] -
                                     as.vector(Y.future[,i]))^2, 1,
                                    mean)))
    rmse.VAR[1,i] <- mean(sqrt(apply((VAR[,,i] -
                                     as.vector(Y.future[,i]))^2, 1,
                                    mean)))

    rmse.norm[1,i] <- mean(sqrt(apply((normonly[,,i] -
                                       as.vector(Y.future[,i]))^2, 1,
                                      mean)))

    # MAE
    mae.MS[1,i] <- mean(abs(MS[,,i] - as.vector(Y.future[,i])))
    mae.VAR[1,i] <- mean(abs(VAR[,,i] - as.vector(Y.future[,i])))
    mae.norm[1,i] <- mean(abs(normonly[,,i] - as.vector(Y.future[,i])))
}

#####################################################
# CRPS
#####################################################
library(verification)

# Compute the CRPS for each forecast set and variable
MS1.crps <- crps(as.vector(Y.future[,1]),
                 cbind(apply(MS[,,1], 2, mean),
                       apply(MS[,,1], 2, sd)))

MS2.crps <- crps(Y.future[,2],
                 cbind(apply(MS[,,2], 2, mean),
                       apply(MS[,,2], 2, sd)))

VAR1.crps <- crps(Y.future[,1],
                  cbind(apply(VAR[,,1], 2, mean),
                        apply(VAR[,,1], 2, sd)))

VAR2.crps <- crps(Y.future[,2],
                  cbind(apply(VAR[,,2], 2, mean),
                        apply(VAR[,,2], 2, sd)))


norm1.crps <- crps(Y.future[,1],
                   cbind(apply(normonly[,,1], 2, mean),
                         apply(normonly[,,1], 2, sd)))

norm2.crps <- crps(Y.future[,2],
                   cbind(apply(normonly[,,1], 2, mean),
                         apply(normonly[,,1], 2, sd)))


# Compute / plot the PITs

# First, used ecdfs to compute the PIT values and then plot.

# Storage for values
MS.pit <- VAR.pit <- norm.pit <- matrix(0, k, 2)

# Do this by variable across the time periods
for(i in 1:m)
{
    tmp <- matrix(as.vector(Y.future[,i]), byrow=TRUE, nrow=N2, ncol=k)
    MS.pit[,i] <- apply(MS[,,i] < tmp, 2, mean)
    VAR.pit[,i] <- apply(VAR[,,i] < tmp, 2, mean)
    norm.pit[,i] <- apply(normonly[,,i] < tmp, 2, mean)
}

# Second, check the PITs for autocorrelation
par(mfrow=c(3,2))
apply(MS.pit, 2, acf)
apply(VAR.pit, 2, acf)
apply(norm.pit, 2, acf)

# Compute the KS tests on the PITS
MS.ks <- apply(MS.pit, 2, ks.test, y="punif")
VAR.ks <- apply(VAR.pit, 2, ks.test, y="punif")
norm.ks <- apply(norm.pit, 2, ks.test, y="punif")

# Third, plot the PITS
pdf(file="toyMS2PIT.pdf")
vnames <- c("Series 1", "Series 2")

par(mfcol=c(3,2), mai=c(0.4,0.25,0.25,0.25), omi=c(0,0.75,0,0))
for(i in 1:m)
{
    hist(MS.pit[,i], main=vnames[i], breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("MSBVAR", side=2, line=3)
    mtext(paste("KS, p = ", round(MS.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)

    hist(VAR.pit[,i], main="", breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("VAR", side=2, line=3)
    mtext(paste("KS, p = ", round(VAR.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)

    hist(norm.pit[,i], main="", breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("Normal", side=2, line=3)
    mtext(paste("KS, p = ", round(norm.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)
}

dev.off()


# Now compute the VRHs
library(ensembleBMA)


par(mfrow=c(3,2))
# Fix this....
MS1.vrh <- verifRankHist(t(MS[,,1]), as.vector(Y.future[,1]))
MS2.vrh <- verifRankHist(t(MS[,,2]), as.vector(Y.future[,2]))
VAR1.vrh <- verifRankHist(t(VAR[,,1]), Y.future[,1])
VAR2.vrh <- verifRankHist(t(VAR[,,2]), Y.future[,2])
norm1.vrh <- verifRankHist(t(normonly[,,1]), Y.future[,1])
norm2.vrh <- verifRankHist(t(normonly[,,2]), Y.future[,2])



# Make a custom plot so we can add our own labels
pdf(file="toyMS2VRH.pdf", width=6, height=8)

par(mfrow=c(3,2), mai=c(0.4,0.25,0.25,0.25), omi=c(0,0.75,0.5,0))

hist(MS1.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
mtext("MSBVAR", side=2, line=3)
mtext("Series 1", side=3, line=1)
tmp <- ks.test(MS1.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(MS2.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
mtext("Series 2", side=3, line=1)
tmp <- ks.test(MS2.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(VAR1.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
mtext("VAR", side=2, line=3)
tmp <- ks.test(VAR1.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(VAR2.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
tmp <- ks.test(VAR2.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(norm1.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
mtext("Normal", side=2, line=3)
tmp <- ks.test(norm1.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(norm2.vrh, breaks=0:1001, prob=TRUE, main="", xlab="")
tmp <- ks.test(norm2.vrh, y="punif", min=0, max=1000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

dev.off()



########################################################
# Plot data and forecasts
########################################################
# Coverage intervals are 90%
VARse <- apply(VAR, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))
normci <- apply(normonly, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))

pdf(file="predictions2.pdf")
par(mfrow=c(1,2))
plot(ts(cbind(rd[,1],t(fout[,,1]),t(VARse[,,1]),t(normci[,,1]))),
     plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
     lty=c(1,2,1,2,2,1,2,2,1,2), ylab="Series 1")

plot(ts(cbind(rd[,2],t(fout[,,2]),t(VARse[,,2]),t(normci[,,2]))),
     plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
     lty=c(1,2,1,2,2,1,2,2,1,2), ylab="Series 2")

dev.off()


# Save the results

save.image(file="toyMS2.RData")


q(save="no")
