# toymix.R
# Toy mixture model to illustrate forecast performance
#
# Patrick T. Brandt
#
#

library(mclust)

# Simulate the mixture model
# Set up parameters
N.s <- 300    # Sample size
N.a <- 50     # Ex ante forecast horizon

x <- sim(modelName="V",
         parameters=list(pro=c(0.5,0.5), mean=c(-1,1),
         variance=list(modelName="VII", d=1, G=2, sigmasq=c(4,4))),
         n=N.s+N.a, set.seed=123)

# Plot the mixture and then overplot with the theoretical input
# densities
pdf(file="mixdgp.pdf", width=4, height=4)
plot(density(x[1:N.s,2]), ylim=c(0,0.5), lwd=2, main="")
xx <- seq(-6,6, by=0.01)
lines(xx, dnorm(xx, mean=-1, sd=sqrt(4)), col="cyan", lwd=2)
lines(xx, dnorm(xx, mean=1, sd=sqrt(4)), col="magenta", lwd=2)

legend(-4, 0.5, legend=c("Mixture, 80:20",
                "N(-1,4)", "N(1,4)"),
       col=c("black", "cyan", "magenta"), lwd=rep(2,3))
dev.off()

# Now estimate and forecast the mixture model
trueModel <- Mclust(x[1:300,2], G=2, modelNames="V")
trueModel1 <- mclustDAtrain(x[1:N.s,2], labels=x[1:N.s,1],
                            G=2, modelNames="V")

# ALT: Do this with a cluster / discrim analysis
## trueModel2 <- mclustDA(train=list(data=x[1:300,2], labels=x[1:300,1]),
##                        test=list(data=x[301:330,2], labels=x[301:330,1]),
##                        pro=c(0.5,0.5), G=2, modelNames="V")


##########################################################################
# FORECASTING PART
#
# USE THREE FORECASTS
# 1) Mixture with different variances to get the fat tails
# 2) Normal linear projection (naive)
# 3) Incorrect N(0,1)



# Mixture forecasts -- very explicit for now: 100 forecasts for 30 cases.
N <- 200
mixforc <- matrix(0, N, N.a)

m1 <- trueModel$parameters$mean[1]
m2 <- trueModel$parameters$mean[2]
sd1 <- sqrt(trueModel$parameters$variance$sigmasq[1])
sd2 <- sqrt(trueModel$parameters$variance$sigmasq[2])
pro1 <- trueModel$parameters$pro[1]
pro2 <- trueModel$parameters$pro[2]

for(i in 1:N)
{
    for(j in 1:ncol(mixforc))
    {
        mixforc[i,j] <- pro1*rnorm(1, mean=m1, sd=sd1) +
                        pro2*rnorm(1, mean=m2, sd=sd2)
    }
}

# Now generate a non-mixture forecast linear projection
nomixforc <- matrix(0, N, N.a)
mall <- mean(x[1:300,2])
sdall <- sd(x[1:300,2])

for(i in 1:N)
{
    nomixforc[i,] <- rnorm(N.a, mean=mall, sd=sdall)
}

# Now an incorrect forecast that is N(0,1)
normforc <- matrix(rnorm(N.a*N), N, N.a)


# Forecast errors computations
xante <- x[(N.s+1):(N.s+N.a),2]

# Average RMSE over the 100 forecasts
rmse.mix <- mean(sqrt(apply(apply(mixforc, 1, "-", xante)^2, 2,
                            mean)))
rmse.nomix <- mean(sqrt(apply(apply(nomixforc, 1, "-", xante)^2, 2,
                              mean)))

rmse.norm <- mean(sqrt(apply(apply(normforc, 1, "-", xante)^2, 2,
                       mean)))

# Average MAE over the 100 forecasts
mae.mix <- mean(apply(apply(apply(mixforc, 1, "-", xante), 2, abs), 2,
                      mean))
mae.nomix <- mean(apply(apply(apply(nomixforc, 1, "-", xante), 2, abs), 2,
                      mean))
mae.norm <- mean(apply(apply(apply(normforc, 1, "-", xante), 2, abs),
                       2, mean))

# Now apply the forecast variation methods from verification pkg
library(verification)

# Compute the CRPS for each forecast set
mix.crps <- crps(xante, c(mean(as.vector(mixforc)),
                          sd(as.vector(mixforc))))

nomix.crps <- crps(xante, c(mean(as.vector(nomixforc)),
                            sd(as.vector(nomixforc))))

norm.crps <- crps(xante, c(mean(as.vector(normforc)),
                           sd(as.vector(normforc))))

# Now compute the VRHs
library(ensembleBMA)
mix.vrh <- verifRank(t(mixforc), xante)
nomix.vrh <- verifRank(t(nomixforc), xante)
norm.vrh <- verifRank(t(normforc), xante)


# Compute / plot the PITs
pdf(file="mixexample.pdf", width=6, height=5)

par(mfrow=c(2,3))
hist(mix.crps$pit, breaks=seq(0,1,by=0.1),
     main="PIT from mixture forecast",
     xlab="Predicted Probability",
     prob=TRUE, ylim=c(0,4))
hist(nomix.crps$pit, breaks=seq(0,1,by=0.1),
     main="PIT from naive forecast",
     xlab="Predicted Probability",
     prob=TRUE, ylim=c(0,4))
hist(norm.crps$pit, breaks=seq(0,1,by=0.1),
     main="PIT from N(0,1) forecast",
     xlab="Predicted Probability",
     prob=TRUE, ylim=c(0,4))

# Compute / plot the VRHs

hist(mix.vrh,
     breaks=10,
     xlim=c(0,220),
     main="VRH from mixture forecast",
     xlab="Rank of observations in ensemble")

hist(nomix.vrh,
     breaks=10,
     xlim=c(0,220),
     main="VRH from naive forecast",
     xlab="Rank of observation in ensemble")

hist(norm.vrh,
     breaks=10,
     xlim=c(0,220),
     main="VRH from N(0,1) forecast",
     xlab="Rank of observation in ensemble")

dev.off()


# Now make a table of the summary fit measures
library(xtable)
out <- rbind(c(rmse.mix, rmse.nomix, rmse.norm),
             c(mae.mix, mae.nomix, mae.norm),
             c(mix.crps$CRPS, nomix.crps$CRPS, norm.crps$CRPS),
             c(mix.crps$IGN, nomix.crps$IGN, norm.crps$IGN))

print(xtable(out))
